@@ -74,12 +74,25 @@ REP_META_CONTEXT = 4
 REPLY_FLAG_DONE = (1 << 0)
 
 # Flags for base:allocation meta context.
+
+# This range does not allocate any data on storage. Examples are a hole
+# in raw image, or zero cluster in qcow2 image. This is flag is optional
+# and may be ommited by a NBD server. Cannot be used to detect
+# unallocated areas in qcow2 exposing data from the backing file.
 STATE_HOLE = (1 << 0)
+
+# The range is read as zero.
 STATE_ZERO = (1 << 1)
 
 # Flags for qemu:dirty-bitmap meta context.
 STATE_DIRTY = (1 << 0)
 
+# Adusted to support merging allocation and dirty bits.
+EXTENT_DIRTY = (1 << 2)
+
+# Extent does not exist, exposing data from the backing file.
+EXTENT_BACKING = (1 << 3)
+
 # Command flags
 CMD_FLAG_NO_HOLE = (1 << 1)
 
@@ -96,6 +109,11 @@ REPLY_TYPE_ERROR_OFFSET = REPLY_ERROR_BASE + 2
 INFO_EXPORT = 0
 INFO_BLOCK_SIZE = 3
 
+# Meta context names.
+BASE_ALLOCATION = "base:allocation"
+QEMU_ALLOCATION_DEPTH = "qemu:allocation-depth"
+QEMU_DIRTY_BITMAP = "qemu:dirty-bitmap:"
+
 # Error replies
 ERR_BASE = 2**31
 REP_ERR_UNSUP = ERR_BASE + 1
@@ -349,7 +367,7 @@ def _parse_url(url):
 
 CONNECTING = 0
 HANDSHAKE = 1
-TRASMISSION = 2
+TRANSMISSION = 2
 CLOSED = 3
 
 
@@ -398,8 +416,12 @@ class Client:
         log.debug("Ready for transmission")
 
     @property
-    def base_allocation(self):
-        return "base:allocation" in self._meta_context
+    def has_base_allocation(self):
+        return BASE_ALLOCATION in self._meta_context
+
+    @property
+    def has_allocation_depth(self):
+        return QEMU_ALLOCATION_DEPTH in self._meta_context
 
     def read(self, offset, length):
         buf = bytearray(length)
@@ -446,7 +468,7 @@ class Client:
         return cmd.reply
 
     def close(self):
-        if self._state in (HANDSHAKE, TRASMISSION):
+        if self._state in (HANDSHAKE, TRANSMISSION):
             self._soft_disconnect()
         else:
             self._hard_disconnect()
@@ -531,7 +553,7 @@ class Client:
 
         self._negotiate_go_option()
 
-        self._state = TRASMISSION
+        self._state = TRANSMISSION
 
     def _send_client_flags(self, flags):
         log.debug("Sending client flags: %x", flags)
@@ -568,14 +590,14 @@ class Client:
         more than one.
         """
         opt = OPT_LIST_META_CONTEXT
-        data = self._format_meta_context_data("qemu:dirty-bitmap:")
+        data = self._format_meta_context_data(QEMU_DIRTY_BITMAP)
         self._send_option(opt, data)
 
         bitmaps = list(self._iter_meta_context_replies(opt))
 
         if len(bitmaps) == 0:
             log.warning(
-                "Server does not support qemu:dirty-bitmap meta context")
+                "Server does not support %s meta context", QEMU_DIRTY_BITMAP)
             return None
 
         if len(bitmaps) > 1:
@@ -592,7 +614,10 @@ class Client:
         """
         opt = OPT_SET_META_CONTEXT
 
-        queries = ["base:allocation"]
+        # qemu:allocation-depth is required to detect holes in qcow2 images -
+        # unallocated clusters exposing data from the backing chain.
+        # Added in qemu 5.2.0.
+        queries = [BASE_ALLOCATION, QEMU_ALLOCATION_DEPTH]
         if dirty_bitmap:
             queries.append(dirty_bitmap)
 
@@ -626,17 +651,14 @@ class Client:
             self._meta_context[ctx_name] = ctx_id
             self._meta_context[ctx_id] = ctx_name
 
-        # Warn if we failed to activate some meta context. This is a
-        # performance issue.
-
-        if "base:allocation" not in self._meta_context:
-            log.warning("Failed to activate base:allocation")
-
-        if dirty_bitmap:
-            if dirty_bitmap in self._meta_context:
+            if ctx_name == dirty_bitmap:
                 self.dirty_bitmap = dirty_bitmap
-            else:
-                log.warning("Failed to activate %s", dirty_bitmap)
+
+        # Warn if we failed to activate some meta context. This may affect
+        # performance or reduce funcionaliity.
+        for ctx_name in queries:
+            if ctx_name not in self._meta_context:
+                log.warning("Failed to activate %s", ctx_name)
 
     def _format_meta_context_data(self, *queries):
         """
@@ -890,13 +912,13 @@ class Client:
         try:
             if self._state == HANDSHAKE:
                 self._send_option(OPT_ABORT)
-            elif self._state == TRASMISSION:
+            elif self._state == TRANSMISSION:
                 cmd = Disc(self._next_handle())
                 self._send_command(cmd)
             else:
                 raise AssertionError(
                     "Cannot initiate soft disconnect at state {!r}"
-                    .foramt(self._state))
+                    .format(self._state))
         except socket.error as e:
             log.debug("Error initiating soft disconnect: %s", e)
         except Exception:
@@ -1052,8 +1074,15 @@ class Client:
             raise ProtocolError(
                 "Received unexpected metadata context id {}".format(ctx_id))
 
+        if ctx_name == self.dirty_bitmap:
+            context = Extent.DIRTY
+        elif ctx_name == QEMU_ALLOCATION_DEPTH:
+            context = Extent.DEPTH
+        else:
+            context = Extent.ALLOC
+
         extents = []
-        for ext in self._recv_extents(length - ctx_id_size):
+        for ext in self._recv_extents(length - ctx_id_size, context):
             if ext.length % self.minimum_block_size:
                 raise ProtocolError(
                     "Invalid extent length {}: not an integer multiple "
@@ -1063,7 +1092,7 @@ class Client:
 
         cmd.reply[ctx_name] = extents
 
-    def _recv_extents(self, length):
+    def _recv_extents(self, length, context):
         """
         Iterator receiving and unpacking extents descriptors form block status
         payload.
@@ -1086,7 +1115,7 @@ class Client:
             # Unpack extents in buffer.
             view = memoryview(buf)
             while len(view):
-                yield Extent.unpack(view[:Extent.size])
+                yield Extent.unpack(view[:Extent.size], context)
                 view = view[Extent.size:]
 
     def _handle_none_chunk(self, flags, length):
@@ -1335,6 +1364,11 @@ class Extent:
     Since this class is mutable, it must not implement __hash__.
     """
 
+    # Contexts ids copied from qemu-nbd.
+    ALLOC = 0  # base:allocation
+    DEPTH = 1  # qemu:allocation-depth
+    DIRTY = 2  # qemu:dirty-bitmap:name
+
     __slots__ = ("length", "_flags")
 
     # 32 bits, length of the extent to which the status below applies
@@ -1349,14 +1383,36 @@ class Extent:
         self._flags = flags
 
     @classmethod
-    def unpack(cls, data):
+    def unpack(cls, data, context=ALLOC):
+        """
+        Create extent instance from extent data.
+
+        Based on context, we map NBD flags bits to private bits to allow
+        merging different types of extents.
+        """
         length, flags = cls.wire_format.unpack(data)
         if length == 0:
             raise ProtocolError(
                 "Invalid extent length=0 flags={}".format(flags))
 
+        log.debug("Extent length=%s flags=%s context=%s",
+                  length, flags, context)
+
+        if context == cls.ALLOC:
+            # The remainder of the flags field is reserved. Servers SHOULD set
+            # it to all-zero; clients MUST ignore unknown flags.
+            flags = flags & (STATE_HOLE | STATE_ZERO)
+        elif context == cls.DIRTY:
+            flags = EXTENT_DIRTY if flags & STATE_DIRTY else 0
+        elif context == cls.DEPTH:
+            flags = EXTENT_BACKING if flags == 0 else 0
+
         return cls(length, flags)
 
+    @classmethod
+    def pack(cls, length, flags):
+        return cls.wire_format.pack(length, flags)
+
     @property
     def flags(self):
         return self._flags
@@ -1370,7 +1426,22 @@ class Extent:
 
     @property
     def hole(self):
-        return bool(self._flags & STATE_HOLE)
+        """
+        Return True if this is a non existing extent in qcow2 image. If the
+        image has a backing file the content of the extent are read from the
+        backing file.
+
+        Return False if the extent exist in the one of the layers, or if the
+        image is a raw image, even if this is unallocated area in a raw
+        image.
+
+        By definition, a hole has also the STATE_HOLE | STATE_ZERO
+        flags. But the oposite is not true.
+
+        For more info see
+        https://lists.nongnu.org/archive/html/qemu-block/2021-06/msg00756.html
+        """
+        return bool(self._flags & EXTENT_BACKING)
 
     @property
     def dirty(self):
@@ -1378,16 +1449,12 @@ class Extent:
         For qemu:dirty-bitmap extent, True if extents was modified and is
         included in this incremental backup.
         """
-        return bool(self._flags & STATE_DIRTY)
+        return bool(self._flags & EXTENT_DIRTY)
 
     def __eq__(self, other):
         return (self.__class__ == other.__class__ and
                 self.length == other.length and
                 self.flags == other.flags)
 
-    # TODO: For python 2; remove when we require python 3.
-    def __ne__(self, other):
-        return not self.__eq__(other)
-
     def __repr__(self):
         return "Extent(length={}, flags={})".format(self.length, self._flags)
