@@ -6,8 +6,6 @@
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 
-from __future__ import absolute_import
-
 import errno
 import os
 import subprocess
@@ -19,10 +17,6 @@ import pytest
 from ovirt_imageio._internal import ioutil
 from ovirt_imageio._internal import util
 
-from . marks import requires_python3
-
-pytestmark = requires_python3
-
 BLOCKSIZE = 4096
 
 requires_root = pytest.mark.skipif(os.geteuid() != 0, reason="Requires root")
