@@ -6,16 +6,13 @@
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 
-from __future__ import absolute_import
-from __future__ import print_function
-
+import http.client as http_client
 import io
 import json
 import os
 import stat
 
 import pytest
-from six.moves import http_client
 
 from ovirt_imageio._internal import config
 from ovirt_imageio._internal import server
@@ -23,10 +20,6 @@ from ovirt_imageio._internal import server
 from . import http
 from . import testutil
 
-from . marks import requires_python3
-
-pytestmark = requires_python3
-
 
 @pytest.fixture(scope="module")
 def srv():
