@@ -29,7 +29,6 @@ from . import testutil
 
 from . marks import (
     flaky_in_ovirt_ci,
-    requires_advanced_virt,
     requires_ipv6,
 )
 
@@ -112,7 +111,9 @@ def test_handshake(tmpdir, export, fmt):
             assert c.minimum_block_size == 1
             assert c.preferred_block_size == 4096
             assert c.maximum_block_size == 32 * 1024**2
-            assert c.base_allocation
+            # Both available in in current qemu-nbd (>= 5.2.0).
+            assert c.has_base_allocation
+            assert c.has_allocation_depth
 
 
 def test_raw_read(tmpdir):
@@ -322,17 +323,21 @@ def test_open_tcp(tmpdir, url_template, host, export):
             assert c.export_size == 1024**3
 
 
-FMT_ZERO_FLAGS_PARAMS = [
-    # raw format does not report holes, only zero status.
-    ("raw", nbd.STATE_ZERO),
-    # qcow2 format report unallocated clusters as holes.
-    ("qcow2", nbd.STATE_ZERO | nbd.STATE_HOLE),
+FMT_HOLE_FLAGS_PARAMS = [
+    # In qemu-nbd 5.2.0 we get only the zero bit.
+    ["raw", nbd.STATE_ZERO],
+    ["qcow2", nbd.STATE_ZERO | nbd.STATE_HOLE | nbd.EXTENT_BACKING],
 ]
 
+# Since qemu-nbd 6.0.0 we get also the hole bit, but we don't get the backing
+# bit, so unallocated areas in raw images are not considered as holes.
+if qemu_nbd.version() >= (6, 0, 0):
+    FMT_HOLE_FLAGS_PARAMS[0][1] |= nbd.STATE_HOLE
 
-@pytest.mark.parametrize("fmt,zero_flags", FMT_ZERO_FLAGS_PARAMS)
-def test_base_allocation_empty(nbd_server, user_file, fmt, zero_flags):
-    size = nbd.MAX_LENGTH
+
+@pytest.mark.parametrize("fmt,hole_flags", FMT_HOLE_FLAGS_PARAMS)
+def test_base_allocation_empty(nbd_server, user_file, fmt, hole_flags):
+    size = 1024**3
     create_image(user_file.path, fmt, size)
 
     nbd_server.image = user_file.path
@@ -341,32 +346,32 @@ def test_base_allocation_empty(nbd_server, user_file, fmt, zero_flags):
 
     with nbd.open(nbd_server.url) as c:
         # Entire image.
-        extents = c.extents(0, size)["base:allocation"]
-        assert extents == [nbd.Extent(length=size, flags=zero_flags)]
+        extents = list(nbdutil.extents(c))
+        assert extents == [nbd.Extent(length=size, flags=hole_flags)]
 
         # First block.
-        extents = c.extents(0, 4096)["base:allocation"]
-        assert extents == [nbd.Extent(length=4096, flags=zero_flags)]
+        extents = list(nbdutil.extents(c, length=4096))
+        assert extents == [nbd.Extent(length=4096, flags=hole_flags)]
 
         # Last block.
-        extents = c.extents(size - 4096, 4096)["base:allocation"]
-        assert extents == [nbd.Extent(length=4096, flags=zero_flags)]
+        extents = list(nbdutil.extents(c, offset=size-4096, length=4096))
+        assert extents == [nbd.Extent(length=4096, flags=hole_flags)]
 
         # Some block.
-        extents = c.extents(4096, 4096)["base:allocation"]
-        assert extents == [nbd.Extent(length=4096, flags=zero_flags)]
+        extents = list(nbdutil.extents(c, offset=4096, length=4096))
+        assert extents == [nbd.Extent(length=4096, flags=hole_flags)]
 
         # Unaligned start.
-        extents = c.extents(4096 - 1, 4096 + 1)["base:allocation"]
-        assert extents == [nbd.Extent(length=4096 + 1, flags=zero_flags)]
+        extents = list(nbdutil.extents(c, offset=4096 - 1, length=4096 + 1))
+        assert extents == [nbd.Extent(length=4096 + 1, flags=hole_flags)]
 
         # Unaligned end.
-        extents = c.extents(4096, 4096 + 1)["base:allocation"]
-        assert extents == [nbd.Extent(length=4096 + 1, flags=zero_flags)]
+        extents = list(nbdutil.extents(c, offset=4096, length=4096 + 1))
+        assert extents == [nbd.Extent(length=4096 + 1, flags=hole_flags)]
 
         # Unaligned start and end.
-        extents = c.extents(4096 - 1, 4096 + 2)["base:allocation"]
-        assert extents == [nbd.Extent(length=4096 + 2, flags=zero_flags)]
+        extents = list(nbdutil.extents(c, offset=4096 - 1, length=4096 + 2))
+        assert extents == [nbd.Extent(length=4096 + 2, flags=hole_flags)]
 
 
 @pytest.mark.parametrize("fmt", ["raw", "qcow2"])
@@ -382,36 +387,36 @@ def test_base_allocation_full(nbd_server, user_file, fmt):
         c.write(0, b"x" * size)
 
         # Entire image.
-        extents = c.extents(0, size)["base:allocation"]
+        extents = list(nbdutil.extents(c))
         assert extents == [nbd.Extent(length=size, flags=0)]
 
         # First block.
-        extents = c.extents(0, 4096)["base:allocation"]
+        extents = list(nbdutil.extents(c, length=4096))
         assert extents == [nbd.Extent(length=4096, flags=0)]
 
         # Last block.
-        extents = c.extents(size - 4096, 4096)["base:allocation"]
+        extents = list(nbdutil.extents(c, offset=size-4096, length=4096))
         assert extents == [nbd.Extent(length=4096, flags=0)]
 
         # Some block.
-        extents = c.extents(4096, 4096)["base:allocation"]
+        extents = list(nbdutil.extents(c, offset=4096, length=4096))
         assert extents == [nbd.Extent(length=4096, flags=0)]
 
         # Unaligned start.
-        extents = c.extents(4096 - 1, 4096 + 1)["base:allocation"]
+        extents = list(nbdutil.extents(c, offset=4096 - 1, length=4096 + 1))
         assert extents == [nbd.Extent(length=4096 + 1, flags=0)]
 
         # Unaligned end.
-        extents = c.extents(4096, 4096 + 1)["base:allocation"]
+        extents = list(nbdutil.extents(c, offset=4096, length=4096 + 1))
         assert extents == [nbd.Extent(length=4096 + 1, flags=0)]
 
         # Unaligned start and end.
-        extents = c.extents(4096 - 1, 4096 + 2)["base:allocation"]
+        extents = list(nbdutil.extents(c, offset=4096 - 1, length=4096 + 2))
         assert extents == [nbd.Extent(length=4096 + 2, flags=0)]
 
 
-@pytest.mark.parametrize("fmt,zero_flags", FMT_ZERO_FLAGS_PARAMS)
-def test_base_allocation_some_data(nbd_server, user_file, fmt, zero_flags):
+@pytest.mark.parametrize("fmt,hole_flags", FMT_HOLE_FLAGS_PARAMS)
+def test_base_allocation_some_data(nbd_server, user_file, fmt, hole_flags):
     size = 1024**3
     create_image(user_file.path, fmt, size)
 
@@ -433,15 +438,15 @@ def test_base_allocation_some_data(nbd_server, user_file, fmt, zero_flags):
 
     assert extents == [
         nbd.Extent(length=data_length, flags=0),
-        nbd.Extent(length=zero_length, flags=zero_flags),
+        nbd.Extent(length=zero_length, flags=hole_flags),
         nbd.Extent(length=data_length, flags=0),
-        nbd.Extent(length=zero_length, flags=zero_flags),
+        nbd.Extent(length=zero_length, flags=hole_flags),
     ]
 
 
-@pytest.mark.parametrize("fmt,zero_flags", FMT_ZERO_FLAGS_PARAMS)
+@pytest.mark.parametrize("fmt,hole_flags", FMT_HOLE_FLAGS_PARAMS)
 def test_base_allocation_some_data_unaligned(
-        nbd_server, user_file, fmt, zero_flags):
+        nbd_server, user_file, fmt, hole_flags):
     size = 1024**2
     create_image(user_file.path, fmt, size)
 
@@ -459,9 +464,9 @@ def test_base_allocation_some_data_unaligned(
         # Unaligned part from first extent and last extent.
         extents = list(nbdutil.extents(c, data_offset - 1, data_length + 2))
         assert extents == [
-            nbd.Extent(length=1, flags=zero_flags),
+            nbd.Extent(length=1, flags=hole_flags),
             nbd.Extent(length=data_length, flags=0),
-            nbd.Extent(length=1, flags=zero_flags),
+            nbd.Extent(length=1, flags=hole_flags),
         ]
 
         # Unaligned part from second extent.
@@ -474,7 +479,7 @@ def test_base_allocation_some_data_unaligned(
         extents = list(nbdutil.extents(c, data_offset + 1, data_length))
         assert extents == [
             nbd.Extent(length=data_length - 1, flags=0),
-            nbd.Extent(length=1, flags=zero_flags),
+            nbd.Extent(length=1, flags=hole_flags),
         ]
 
 
@@ -549,7 +554,9 @@ def test_full_backup_handshake(tmpdir, fmt, nbd_sock):
             assert c.minimum_block_size == 1
             assert c.preferred_block_size == 4096
             assert c.maximum_block_size == 32 * 1024**2
-            assert c.base_allocation
+            # Both available in in current qemu (>= 5.2.0).
+            assert c.has_base_allocation
+            assert c.has_allocation_depth
 
 
 @pytest.mark.parametrize("fmt", [
@@ -559,7 +566,7 @@ def test_full_backup_handshake(tmpdir, fmt, nbd_sock):
             distro.is_centos("8") and ci.is_ovirt(),
             reason="unaligned write fails on el8/oVirt CI")
     ),
-    pytest.param("qcow2", marks=requires_advanced_virt),
+    pytest.param("qcow2"),
 ])
 def test_full_backup_single_image(tmpdir, user_file, fmt, nbd_sock):
     chunk_size = 1024**3
@@ -589,11 +596,7 @@ def test_full_backup_single_image(tmpdir, user_file, fmt, nbd_sock):
 
 @pytest.mark.parametrize("checkpoint", [
     pytest.param(None, id="no-checkpoint"),
-    pytest.param(
-        "check1",
-        id="with-checkpoint",
-        marks=requires_advanced_virt
-    ),
+    pytest.param("check1", id="with-checkpoint"),
 ])
 def test_full_backup_complete_chain(tmpdir, nbd_sock, checkpoint):
     depth = 3
@@ -647,35 +650,165 @@ def verify_full_backup(sock, export_name):
 
 
 def test_extent_base_allocation():
-    # Allocated aread with data.
-    ext = nbd.Extent(4096, 0)
+    # Allocated area with data.
+    data = nbd.Extent.pack(4096, 0)
+    ext = nbd.Extent.unpack(data)
     assert not ext.zero
     assert not ext.hole
     assert ext.flags == 0
 
-    # Allocated aread that reads as zero.
-    ext = nbd.Extent(4096, nbd.STATE_ZERO)
+    # Allocated area that reads as zero.
+    data = nbd.Extent.pack(4096, nbd.STATE_ZERO)
+    ext = nbd.Extent.unpack(data)
     assert ext.zero
     assert not ext.hole
     assert ext.flags == nbd.STATE_ZERO
 
-    # Unallocated aread that reads as zero.
-    ext = nbd.Extent(4096, nbd.STATE_ZERO | nbd.STATE_HOLE)
+    # Unallocated area that reads as zero.
+    data = nbd.Extent.pack(4096, nbd.STATE_ZERO | nbd.STATE_HOLE)
+    ext = nbd.Extent.unpack(data)
     assert ext.zero
-    assert ext.hole
+    # Hole can be detected only when using depth data with qcow2 image.
+    assert not ext.hole
     assert ext.flags == nbd.STATE_ZERO | nbd.STATE_HOLE
 
+    # Ignore unexpected bits
+    data = nbd.Extent.pack(4096, 0xffff)
+    ext = nbd.Extent.unpack(data)
+    assert ext.zero
+    assert not ext.hole
+    assert not ext.dirty
+    assert ext.flags == nbd.STATE_HOLE | nbd.STATE_ZERO
+
 
 def test_extent_dirty_bitmap():
     # Clean area.
-    ext = nbd.Extent(4096, 0)
+    data = nbd.Extent.pack(4096, 0)
+    ext = nbd.Extent.unpack(data, nbd.Extent.DIRTY)
     assert not ext.dirty
     assert ext.flags == 0
 
     # Dirty area.
-    ext = nbd.Extent(4096, nbd.STATE_DIRTY)
+    data = nbd.Extent.pack(4096, nbd.STATE_DIRTY)
+    ext = nbd.Extent.unpack(data, nbd.Extent.DIRTY)
     assert ext.dirty
-    assert ext.flags == nbd.STATE_DIRTY
+    # nbd.STATE_DIRTY is mapped to nbd.EXTENT_DIRTY to allow merging allocation
+    # and dirty bitmap info in same extent.
+    assert ext.flags == nbd.EXTENT_DIRTY
+
+    # Ignore unexpected bits
+    data = nbd.Extent.pack(4096, 0xffff)
+    ext = nbd.Extent.unpack(data, nbd.Extent.DIRTY)
+    assert not ext.zero
+    assert not ext.hole
+    assert ext.dirty
+    assert ext.flags == nbd.EXTENT_DIRTY
+
+
+def test_extent_depth():
+    # Unallocated extent.
+    depth_data = nbd.Extent.pack(65536, 0)
+    ext = nbd.Extent.unpack(depth_data, nbd.Extent.DEPTH)
+    assert ext.length == 65536
+    assert not ext.zero
+    # Hole status is determined only by the EXTENT_BACKING bit.
+    assert ext.hole
+    assert not ext.dirty
+    assert ext.flags == nbd.EXTENT_BACKING
+
+    # Allocated extents from top layer.
+    data = nbd.Extent.pack(4096, 1)
+    ext = nbd.Extent.unpack(data, nbd.Extent.DEPTH)
+    assert not ext.zero
+    assert not ext.hole
+    assert not ext.dirty
+    assert ext.flags == 0
+
+    # Allocated extents from top backing file.
+    data = nbd.Extent.pack(4096, 2)
+    ext = nbd.Extent.unpack(data, nbd.Extent.DEPTH)
+    assert not ext.zero
+    assert not ext.hole
+    assert not ext.dirty
+    assert ext.flags == 0
+
+
+def test_extent_compare():
+    assert nbd.Extent(4096, 0) == nbd.Extent(4096, 0)
+    assert nbd.Extent(4096, 0) != nbd.Extent(4096, nbd.STATE_ZERO)
+
+
+def test_extent_merge_clean_hole():
+    alloc_data = nbd.Extent.pack(4096, nbd.STATE_ZERO | nbd.STATE_HOLE)
+    alloc = nbd.Extent.unpack(alloc_data)
+
+    dirty_data = nbd.Extent.pack(4096, 0)
+    dirty = nbd.Extent.unpack(dirty_data, nbd.Extent.DIRTY)
+
+    merged = nbd.Extent(4096, alloc.flags | dirty.flags)
+
+    assert merged.length == 4096
+    assert merged.zero is True
+    # Hole can be detected only when using depth data with qcow2 image.
+    assert merged.hole is False
+    assert not merged.dirty
+
+
+def test_extent_merge_dirty_data():
+    alloc_data = nbd.Extent.pack(4096, 0)
+    alloc = nbd.Extent.unpack(alloc_data, nbd.Extent.DIRTY)
+
+    dirty_data = nbd.Extent.pack(4096, nbd.STATE_DIRTY)
+    dirty = nbd.Extent.unpack(dirty_data, nbd.Extent.DIRTY)
+
+    merged = nbd.Extent(4096, alloc.flags | dirty.flags)
+
+    assert merged.length == 4096
+    assert not merged.zero
+    assert not merged.hole
+    assert merged.dirty
+
+
+def test_extent_merge_qcow2_unallocated_cluster():
+    alloc_data = nbd.Extent.pack(65536, nbd.STATE_ZERO | nbd.STATE_HOLE)
+    alloc = nbd.Extent.unpack(alloc_data)
+
+    depth_data = nbd.Extent.pack(65536, 0)
+    depth = nbd.Extent.unpack(depth_data, nbd.Extent.DEPTH)
+
+    merged = nbd.Extent(65536, alloc.flags | depth.flags)
+
+    assert merged.length == 65536
+    assert merged.zero is True
+    assert merged.hole is True
+
+
+def test_extent_merge_qcow2_zero_cluster():
+    alloc_data = nbd.Extent.pack(65536, nbd.STATE_ZERO | nbd.STATE_HOLE)
+    alloc = nbd.Extent.unpack(alloc_data)
+
+    depth_data = nbd.Extent.pack(65536, 1)
+    depth = nbd.Extent.unpack(depth_data, nbd.Extent.DEPTH)
+
+    merged = nbd.Extent(65536, alloc.flags | depth.flags)
+
+    assert merged.length == 65536
+    assert merged.zero is True
+    assert merged.hole is False
+
+
+def test_extent_merge_qcow2_data_cluster():
+    alloc_data = nbd.Extent.pack(65536, 0)
+    alloc = nbd.Extent.unpack(alloc_data)
+
+    depth_data = nbd.Extent.pack(65536, 1)
+    depth = nbd.Extent.unpack(depth_data, nbd.Extent.DEPTH)
+
+    merged = nbd.Extent(65536, alloc.flags | depth.flags)
+
+    assert merged.length == 65536
+    assert merged.zero is False
+    assert merged.hole is False
 
 
 def test_reply_error_structured():
