@@ -21,7 +21,7 @@ from ovirt_imageio._internal import qemu_nbd
 from ovirt_imageio._internal import sockutil
 
 from . import testutil
-from . marks import flaky_in_ovirt_ci, requires_advanced_virt
+from . marks import flaky_in_ovirt_ci
 
 
 @pytest.mark.parametrize("addr,export,url", [
@@ -271,7 +271,6 @@ def test_shared(tmpdir, fmt):
     qemu_img.compare(src, dst)
 
 
-@requires_advanced_virt
 def test_dirty_bitmap(tmpdir):
     size = 1024**2
 
@@ -289,11 +288,11 @@ def test_dirty_bitmap(tmpdir):
 
     # Read dirty extents.
     with qemu_nbd.open(img, "qcow2", read_only=True, bitmap="b0") as c:
-        extents = c.extents(0, size)["qemu:dirty-bitmap:b0"]
+        extents = c.extents(0, size)[nbd.QEMU_DIRTY_BITMAP + "b0"]
 
     bitmap = qemu_img.info(img)["format-specific"]["data"]["bitmaps"][0]
 
     assert extents == [
-        nbd.Extent(length=bitmap["granularity"], flags=1),
+        nbd.Extent(length=bitmap["granularity"], flags=nbd.EXTENT_DIRTY),
         nbd.Extent(length=size - bitmap["granularity"], flags=0),
     ]
